/* Entities.cpp
	This class contains all the information about the Entity class and its derived classes.
	Entities are units which take up a spot on the grid, and also Render and Update via inheritence
	from the Renderer class.
	All the units' behaviours are contained in this module, however, some functions from Game
	are used to interact with the game world.
*/

#include "Entities.h"
#include <Math.h>

/////////////////////////// Entity

// Default Constructor
Entity::Entity() {
	state = STATE_ORDERS;
	lorderType = LO_NONE;
	team = TEAM_NONE;
	target = 0;
	localOrderTimeLeft = 0;
}

// ApplyState, called mainly to change the color of the entity based on their state and clear variables.
void Entity::ApplyState(EntityState state) {
	this->state = state;

	// Clear local orders if we are not in the local order state.
	if(state != STATE_MOVINGTOLOCALORDER) {
		lorderType = LO_NONE;
		localOrderTimeLeft = 0;
		localOrder = Vec2(-1, -1);
	}

	if(state == STATE_ORDERS || state == STATE_MOVINGTOLOCALORDER) {
		col = Game::Instance()->GetColorOfTeam(team) - Color(128, 128, 128);
	}
	else if(state == STATE_ENEMYCONTACT || state == STATE_LOADINGWEAPON) {
		lorderType = LO_NONE;
		col = Game::Instance()->GetColorOfTeam(team) + Color(64, 64, 64);
	}
	else {
		col = Color(255,255,255);
	}
}

// The base Entity Update function. Manages death and calling the overridden "Act" every [speed] frames.
void Entity::Update() {
	
	if(hp <= 0) {
		Game::Instance()->DeleteObject(this);	// Queue this entity for deletion at the end of this cycle
	}
	else {
		moveWait--;
		if(moveWait <= 0) {
			moveWait = speed;
			Act();
		}
	}
}

// Takes a distance, and finds the nearest enemy in this distance. The distance is calculated as an approximate
// "grid circle", and is calculated starting at the inner "ring" around the unit, and moves outwards, stopping
// when it finds a potential target.
Entity* Entity::GetNearestEnemy(int dist) {
	Vec2 mypos = Vec2(this->x,this->y);

	dist = dist + 1; // convert from hill building to range check

	for(int j = dist - 1; j > 0; j--) {
		for(int k = 0; k < dist - j; k++) {

			// Our four quadrants
			Vec2 q[4];
			q[0] = Vec2(k, dist - j - k);
			q[1] = Vec2(-q[0].y, q[0].x);
			q[2] = Vec2(-q[0].x, -q[0].y);
			q[3] = Vec2(-q[1].x, -q[1].y);
			
			// May favor enemies in a certain quadrant (q0)
			for(int jj = 0; jj < 4; jj++) {
				Vec2 realPos = Vec2(mypos.x + q[jj].x, mypos.y + q[jj].y);
					Entity* curr = Game::Instance()->GetGridAt(realPos);	
			
					// If we have any target, take it and return, since it is (tied) for the closest
					if(curr && curr->team != this->team) {
						return curr;
					}
			}
		}
	}

	return 0;
}

// Move on the grid, updating the game's grid as well as local x/y vars
void Entity::SetPos(Vec2 pos) {
	Game::Instance()->EntityMoved(Vec2(x, y), pos, this);
	x = pos.x;
	y = pos.y;
}

// Finds the best move, rating first on distance to pos, then on the quality of the terrain.
Vec2 Entity::GetClosestAdj(Vec2 pos) {
	float best = 999999;
	Vec2 closest = Vec2(x,y);
	float mydist = closest.distSqr(pos);
	int bestTerrain = 0;

	for(int i = DIR_UPLEFT; i <= DIR_NONE; i++) {
		Vec2 test = Vec2(x,y) + Game::Instance()->DirToVec2(Direction(i));
		if(test.x < 0 || test.y < 0 || test.x >= CELLS_X || test.y >= CELLS_Y)
			continue;
		if(CanMove(test)) {
			float dist = test.distSqr(pos);
			if(dist < best) {
				bestTerrain = Game::Instance()->GetTerrainAt(test);
				best = dist;
				closest = test;
			}
			else if(dist == best && Game::Instance()->GetTerrainAt(test) > bestTerrain) {
				bestTerrain = Game::Instance()->GetTerrainAt(test);
				best = dist;
				closest = test;
			}
		}
	}	
	return closest; 
}

// This is used when an entity is attacking another. It will move to the best adjacent spot that has the best
// cover and is in its weapons range. If the entity has no target, it will merely find the best cover.
Vec2 Entity::GetBestTerrainInRange() {
	Vec2 closest = Vec2(x, y);
	int best = Game::Instance()->GetTerrainAt(closest);

	// Iterate through directions
	for(int i = DIR_UPLEFT; i <= DIR_NONE; i++) {
		Vec2 test = Vec2(x,y) + Game::Instance()->DirToVec2(Direction(i));

		// Position is invalid, or we have a target and we're not in range of it.
		if(!Game::Instance()->InGrid(test) || (this->target && !TargetInRange(test.x, test.y))) 
			continue;
		if(CanMove(test) && Game::Instance()->GetTerrainAt(test) > best) {
			best = Game::Instance()->GetTerrainAt(test);
			closest = test;
		}
	}
	return closest;
}

// Entity move by direction rather than absolute position
void Entity::MoveTo(Direction dir) {

	Vec2 dirvec = Game::Instance()->DirToVec2(dir);
	Vec2 target = Vec2(x, y) + dirvec;
	
	if(CanMove(target)) {
		SetPos(target);
	}

}

// Causes [sdmg] damage to all entities within a square of length 2 * [radius] + 1
// centered around point [pos]. Team [immune] is ignored for damage. 
void Entity::DoSplashDmg(Vec2 pos, int sdmg, int radius, Team immune) {

	for(int i = pos.x-radius; i <= pos.x+radius; i++) {
		for(int j = pos.y-radius; j <= pos.y+radius; j++) {
			Entity* curr = Game::Instance()->GetGridAt(Vec2(i ,j));	
			
			if(curr && curr->team != immune) {
				curr->TakeDamage(this, sdmg);
			}
		}
	}
}

// Modifies an Entity's localorder, sets up the state and the variables
void Entity::ApplyLocalOrder(Vec2 pos, LocalOrderType orderType, int timeLeft) {
	localOrder = pos;
	ApplyState(STATE_MOVINGTOLOCALORDER);
	lorderType = orderType;
	localOrderTimeLeft = timeLeft;
}

// Reduces the entity's hp, creates particle effects, and initiates a
// local order to investigate the entity attacking it.
void Entity::TakeDamage(Entity *from, float amount) {
	int terrainval = Game::Instance()->GetTerrainAt(Vec2(x, y));
	float modifier;
	if(terrainval == 0)
		modifier = 1;
	else
		modifier = (30.0 - terrainval) / 30.0;

	hp -= (amount * modifier);
	if(!target) {
		ApplyLocalOrder(Vec2(from->x, from->y), LO_ATTACKED, 10);
	}
	
	if(hp > 0) {
		for(int i = 0; i < 3; i++) {
			Particle *np = new Particle(1, 8 + rand() % 4, Color(200 + rand() % 50, 128 + rand() % 128, 0), 1, 5, Game::Instance()->GetCenterOfCell(Vec2(x, y)), 20);
			Game::Instance()->CreateObject(np);
		}
	}
	else {
		for(int i = 0; i < 10; i++) {
			Particle *np = new Particle(2, 8 + rand() % 4, Color(128 + rand() % 128, 0, 0), 1, 5, Game::Instance()->GetCenterOfCell(Vec2(x, y)), 20);
			Game::Instance()->CreateObject(np);
		}
	}
}

// Is the position valid and unobstructed?
bool Entity::CanMove(Vec2 pos) {
	if(pos.x < 0 || pos.y < 0 || pos.x >= CELLS_X || pos.y >= CELLS_Y)
		return false;
	if(Game::Instance()->GetGridAt(pos) == 0)
		return true;
	else
		return false;
}

// The entity class maintains only structs and primitives, should be good for cleanup.
// The pointers are references to memory managed elsewhere.
void Entity::Destroy() {
	
}

// If this entity were at point (fromx, fromy) would it be able to hit its target?
bool Entity::TargetInRange(int fromx, int fromy) {
	return Vec2(fromx, fromy).distSqr(Vec2(target->x, target->y)) <= range * range;
}

// Is this entity's target in range?
bool Entity::TargetInRange() {
	return TargetInRange(this->x, this->y);
}

// Logic for turrets (bases)
void Entity::ActAsTurret() {

	// Find a target
	Entity *nt = GetNearestEnemy(range);
	target = nt;
	if(target) {
		ApplyState(STATE_ENEMYCONTACT);
	}
	else {
		ApplyState(STATE_ORDERS);
	}

	// If we have a target, attack it!
	if(state == STATE_ENEMYCONTACT) {
		Attack();
	}
}

// Logic for soldiers (snipers too!)
void Entity::ActAsSoldier() {

	// Arbitrary amount of time for a soldier to go exploring
	int localOrderTime = 5;

	// Get the closest enemy and enter enemycontact,
	// or if we're in the enemycontact state, go to idle
	Entity *nt = GetNearestEnemy(range);
	if(nt) {
		ApplyState(STATE_ENEMYCONTACT);
	}
	else {
		if(state == STATE_ENEMYCONTACT) {
			ApplyState(STATE_ORDERS);
		}
	}
	// Update the target
	target = nt;

	// Default state, go to global orders
	if(state == STATE_ORDERS) {

		// Are there any allies in trouble?
		Vec2 friendly = Game::Instance()->GetClosestFriendlyTarget(this, 0);
		
		// If so, investigate!
		if(friendly.x >= 0 && friendly.y >= 0) {
			ApplyLocalOrder(friendly, LO_EXPLORING, localOrderTime);
		}
		// If we have global orders, follow them!
		else if(globalOrder.Valid()) {
			Vec2 etarget = GetClosestAdj(globalOrder);
			SetPos(etarget);
		}
		else {
			// No orders given, no enemy contact or friendlies in range, do nothing.
		}
	}
	else if(state == STATE_ENEMYCONTACT) {

		// This is checked above, but better safe then sorry? :/
		if(!target) {
			ApplyState(STATE_ORDERS);
		}
		else {
			// If we have a target, let's see if we can move to better terrain
			Vec2 bestTerrain = GetBestTerrainInRange();
			if(bestTerrain != Vec2(x,y)) 
				SetPos(bestTerrain);

			// Let's hit our target!
			if(TargetInRange())
				Attack();
		}
	}
	else if(state == STATE_MOVINGTOLOCALORDER) {
		/*Vec2 friendly = Game::Instance()->GetClosestFriendlyTarget(this, 0);
		if(friendly.x >= 0 && friendly.y >= 0) {
			ApplyLocalOrder(friendly, LO_EXPLORING, localOrderTime);
		}*/
		// If we have a local order and we still have exploring time left, let's explore
		// some more!
		if(localOrder.Valid() && localOrderTimeLeft != 0) {
			Vec2 ltarget = GetClosestAdj(localOrder);
			SetPos(ltarget);
			if(localOrderTimeLeft > 0)
				localOrderTimeLeft--;
		}
		else {
			lorderType = LO_NONE;
			ApplyState(STATE_ORDERS);
		}
	}
}


//////////////////////////////// TANK

Tank::Tank(){

}

// Constructor for Tank
Tank::Tank(int x, int y, float hp, int speed, int power, Color col, Team team) : Entity(){
	this->x = x;
	this->y = y;
	this->hp = hp;
	this->col = col;
	this->power = power;
	this->team = team;
	this->speed = speed;
	moveWait = speed;

	this->dmg = 10;
	this->range = 10;
	this->loadingTime = -1;
	this->weaponTarget = Vec2(-1, -1);
}

// Draw the tank, and if it's loading, draw a square on it and a line to its target.
void Tank::Render() {
	float size = 1.0f;
	size = (int)(size * power);
	Game *g = Game::Instance();

	Vec2 pos = g->GetCenterOfCell(Vec2(x, y));
	gfx->GFXrectfill(pos.x - size, pos.y - size, pos.x + size, pos.y + size, col);
	if(state == STATE_LOADINGWEAPON) 
		gfx->GFXrectfill(pos.x - size/2, pos.y - size/2, pos.x + size/2, pos.y + size/2, Color(255,255,255));

	if(state == STATE_LOADINGWEAPON && weaponTarget.Valid()) {
		Vec2 tPos = g->GetCenterOfCell(weaponTarget);
		gfx->GFXline(pos.x, pos.y, tPos.x, tPos.y, g->GetColorOfTeam(team) - Color(150, 150, 150));
	}
}

// Act as a tank!
// Tanks have a loading time, which represents a certain number of moves before they are able to fire.
// During this time the tank cannot do anything else.
void Tank::Act() {

	// If we're not in combat or loading the weapon, find a potential new target
	if(state == STATE_ORDERS || state == STATE_MOVINGTOLOCALORDER) {
		Entity *nt = GetNearestEnemy(range);
		if(nt) {
			ApplyState(STATE_ENEMYCONTACT);
			target = nt;
		}
	}

	// Standard global orders state, similar to soldier
	if(state == STATE_ORDERS) {
		Vec2 friendly = Game::Instance()->GetClosestFriendlyTarget(this, 0);
		if(friendly.x >= 0 && friendly.y >= 0) {
			ApplyLocalOrder(friendly, LO_EXPLORING, 5);
		}
		else if(globalOrder.Valid()) {
			Vec2 etarget = GetClosestAdj(globalOrder);
			SetPos(etarget);
		}
		else {

		}
	}
	// If we have a target, let's lock in on it and start loading.
	else if(state == STATE_ENEMYCONTACT) {
		if(!target) {
			ApplyState(STATE_ORDERS);
		}
		else {
			weaponTarget = Vec2(target->x, target->y);
			loadingTime = 5;
			ApplyState(STATE_LOADINGWEAPON);
		}
	}
	// Standard local order exploring stuff, like soldier
	else if(state == STATE_MOVINGTOLOCALORDER) {
		if(localOrder.Valid() && localOrderTimeLeft > 0) {
			Vec2 etarget = GetClosestAdj(localOrder);
			SetPos(etarget);
			if(localOrderTimeLeft > 0)
				localOrderTimeLeft--;
		}
		else {
			ApplyState(STATE_ORDERS);
		}
	}
	// If we still have loading time left, decrement it and continue,
	// otherwise, fire!
	else if(state == STATE_LOADINGWEAPON) {
		if(!weaponTarget.Valid())
			ApplyState(STATE_ORDERS);
		else {
			if(loadingTime == 0) {
				Attack();
				ApplyState(STATE_ORDERS);
				weaponTarget = Vec2(-1, -1);
			}

			loadingTime = loadingTime - 1;
		}
	}

}

// Create an explosion and do some splash damage
void Tank::Attack() {
	//target->TakeDamage(team, dmg); 
	Game::Instance()->CreateExplosion(weaponTarget);
	DoSplashDmg(weaponTarget, 50, 2, team);
}

/////////////////////////// SOLDIER

// Constructor, bases the stats given to the soldier on the SoldierType given.
// there is also a degree of randomness added.
Soldier::Soldier(int x, int y, SoldierType type, Team team) {
	this->team = team;
	this->type = type;

	switch(type) {
		case SOLDIER_SNIPER:
			this->hp = 50;
			this->speed = 25 + rand() % 10;
			this->power = rand() % 5 + 1;
			this->dmg = 10;
			this->range = 20;
		break;
		default:
			this->hp = 100;
			this->speed = 10 + rand() % 10;
			this->power = rand() % 5 + 1;
			this->dmg = 10;
			this->range = 5;
		break;
	}

	moveWait = speed;
	this->x = x;
	this->y = y;
	this->col = Game::Instance()->GetColorOfTeam(team) - Color(128, 128, 128);
	this->laseron = true;
}

// If we want to create a soldier of some other type for some reason. This could probably be
// refactored out.
Soldier::Soldier(int x, int y, float hp, int speed, int power, Color col, Team team) : Entity() {
	//Entity::Entity();
	this->x = x;
	this->y = y;
	this->hp = hp;
	this->col = col;
	this->power = power;
	this->team = team;
	this->speed = speed;
	moveWait = speed;

	this->dmg = 10;
	this->range = 5;
	this->laseron = true;
}

// Draw the soldier (triangle), and if it's a sniper, draw a plus to differentiate them.
void Soldier::Render() {
	float size = 1.0f;
	size = (int)(size * power);
	Game *g = Game::Instance();

	Vec2 pos = g->GetCenterOfCell(Vec2(x, y));
	gfx->GFXtriangle(pos.x - size, pos.y + size, pos.x, pos.y - size, pos.x + size, pos.y + size, col);
	/*if(target) {
		Vec2 tPos = g->GetCenterOfCell(Vec2(target->x, target->y));
		if(laseron || type == SOLDIER_SNIPER) {
			Color col = g->GetColorOfTeam(team);
			gfx->GFXline(pos.x, pos.y, tPos.x, tPos.y, col + Color(200, 200, 200));
		}
	}
	*/
	if(type == SOLDIER_SNIPER) {
		gfx->GFXline(pos.x - 2, pos.y, pos.x + 2, pos.y, Color(255,255,255));
		gfx->GFXline(pos.x , pos.y - 2, pos.x, pos.y + 2, Color(255,255,255));
	}
}

// Soldiers can use the base class Entity's ActAsSoldier method for their logic
void Soldier::Act() {
	ActAsSoldier();
}

// Fire a laser at the target.
void Soldier::Attack() {
	Game *g = Game::Instance();
	//laseron = !laseron;
	g->CreateObject(new Projectile(this, dmg, 2, Color(255,255,0), 5, g->GetCenterOfCell(Vec2(this->x, this->y)), g->GetCenterOfCell(Vec2(target->x, target->y)), PT_LINE));
}


////////////////////// Base

Base::Base(int x, int y, float hp, Team team) : Entity() {
	this->x = x;
	this->y = y;
	this->hp = hp;
	this->team = team;
	this->speed = 5;
	this->range = 25;
	this->dmg = 10;
	this->moveWait = this->speed;
}

// Draw the base
void Base::Render() {
	Game *g = Game::Instance();

	Vec2 pos = g->GetCenterOfCell(Vec2(x, y));
	gfx->GFXrectfill(pos.x - 5, pos.y, pos.x + 5, pos.y + 5, g->GetColorOfTeam(team));
	gfx->GFXtriangle(pos.x - 5, pos.y, pos.x, pos.y - 5, pos.x +5, pos.y, g->GetColorOfTeam(team));
}

// Bases act as turrets
void Base::Act() {
	ActAsTurret();
}

// Almost exactly the same as soldier
void Base::Attack() {
	Game *g = Game::Instance();
	g->CreateObject(new Projectile(this, dmg, 5, Color(255,255,0), 3, g->GetCenterOfCell(Vec2(this->x, this->y)), g->GetCenterOfCell(Vec2(target->x, target->y)), PT_LINE));
}

// Call entity's update function, and if this entity has been killed this frame, notify Game that this base's team
// has been eliminated!
void Base::Update() {
	Entity::Update();
	if(hp <= 0)
		Game::Instance()->TeamEliminated(team);
}