/* Game.cpp
	This class contains all the information about the game world (terrain, entity grid). However, its
	larger purpose is as a manager of Entities, Particles, Renderers, etc. It handles the creation of 
	new Entities and the cleanup of killed enemies. 
	The class is independant of any timing, and assumes a fixed delta time provided by the class calling
	its update/render functions.
	The game class is a singleton, and is globally accessable with a single entry point (Game::Instance())
	The class is also the place where helper functions such as DirToVec2 are placed for this example.
*/


#include "Game.h"
#include "Entities.h"

// Singleton pattern
Game* Game::instance = 0;

Game* Game::Instance() {
	if(instance == 0) {
		instance = new Game();
	}
	return instance;
}

// Update function, called every dt. Updates the logic of all renderers in the game.
void Game::Update() {

	switch(state) {
		case GS_INGAME:
			// Iterate through objects list, update them.
			for(Renderers::iterator ri = objects.begin(); ri != objects.end(); ri++) {
				Renderer *ren = (*ri);
				ren->Update();
			}
			
			// Atm unused, terrain is not changed.
			UpdateTerrain();
		break;
		case GS_GAMEOVER:

		break;
	}

	// Delete all objects in the deletion queue list.
	CleanUp();
}

// If a team has been eliminated, end the game!
// The engine DOES support more than one team, but this iteration of the program was designed
// for two teams.
void Game::TeamEliminated(Team team) {
	state = GS_GAMEOVER;
	loser = team;
}

// Iterate through the deletion queue, find the corresponding Renderer, remove it from the objects list.
// next, call the corresponding delete function based on the result of GetType(). Atm particle or entity.
void Game::CleanUp() {
	for(Renderers::iterator ri = deletionQueue.begin(); ri != deletionQueue.end(); ri++) {
		Renderer *toDelete = (*ri);
		int i = 0;
		if(toDelete) {
			for(Renderers::iterator reni = objects.begin(); reni != objects.end(); reni++) {
				Renderer *curr = (*reni);

				if(curr == toDelete) {
					objects.erase(reni);

					// Since this particular game uses a separate list of entities (to give team orders,
					// check ally or enemy status), they need a special function to remove them from their list.
					switch(curr->GetType()) {
					case RT_BASE:
					case RT_ENTITY:
						DestroyEntity((Entity *)curr);
					break;
					case RT_PARTICLE:
						DestroyParticle((Particle *)curr);
					break;
					}
					
					// Does the object require any specific internal cleanup?
					curr->Destroy();
					break;
				}
			}
			// Actually free the memory.
			delete(toDelete);
			toDelete = 0;
		}
	}
	deletionQueue.clear();
}

// Add an renderer to the object list.
void Game::CreateObject(Renderer *renderer) {
	objects.push_back(renderer);
}

// WARNING: This integrity check has been made optional in the interest of performance. If it becomes a problem, 
// it can be reinstated. All we need to do is any renderer that can be deleted from more than one place needs to have
// any deletes after the first called as safe (or all of them).
// ie if the same object is added to the deletionQueue more than once it'll try to delete the memory twice and
// cause issues.
void Game::DeleteObject(Renderer *obj, bool safe) {
	if(safe) {
		for(Renderers::iterator ri = deletionQueue.begin(); ri != deletionQueue.end(); ri++) {
			Renderer *curr = (*ri);
			if(curr == obj)
				return;
		}
	}
	deletionQueue.push_back(obj);
}

// This is called after the entity has been removed from the objects list. 
// It first removes the entity from the entity list, and clears the targets of
// any entities targetting this one.
// Note: Could probably be optimized by putting target removal candidates into a list
// but it's probably not a huge time saver.
void Game::DestroyEntity(Entity *toDelete) {
	grid[toDelete->x][toDelete->y] = 0;

	for(Entities::iterator enti = entities.begin(); enti != entities.end(); enti++) {
		Entity *curr = (*enti);

		if(curr == toDelete) {
			entities.erase(enti);
			break;
		}
	}
	for(Entities::iterator ei = entities.begin(); ei != entities.end(); ei++) {
		Entity *curr = (*ei);
		if(curr->target == toDelete) {
			curr->target = 0;
		}
	}
}

// Not used, as we don't maintain a list of particles.
void Game::DestroyParticle(Particle *toDelete) {

}

// Static terrain for now.
void Game::UpdateTerrain() {
	
}

// px and py is the screen point
// gx and gy is the grid cell
// Is world point (px, py) in the grid cell (gx, gy)?
bool Game::ScreenPointInGrid(int px, int py, int gx, int gy) {
		int left = gx*CELL_SIZE;
		int up = gy*CELL_SIZE;
		int right = left + CELL_SIZE;
		int down = up + CELL_SIZE;

		return(px >= left && px <= right && py >= up && py <= down); 
}

// Translate a screen point to a grid cell
Vec2 Game::ScreenPointToGrid(int px, int py) {
	return Vec2((int)(px / CELL_SIZE), (int)(py / CELL_SIZE));
}

// Colors of different teams, for reusability
Color Game::GetColorOfTeam(Team team) {
	if(team == TEAM_RED) {
		return(Color(255, 0, 0));
	}
	else if(team == TEAM_BLUE) {
		return(Color(0, 0, 255));
	}
	else if(team == TEAM_GREEN) {
		return(Color(0, 255, 0));
	}
	else if(team == TEAM_YELLOW) {
		return(Color(255, 255, 0));
	}
	else
		return(Color(255, 255, 255));
}

// If the grid at the given point is free, add a new entity to the game
Entity* Game::AddEntity(Entity *newEntity) {
	int x = newEntity->x;
	int y = newEntity->y;

	// If pos is invalid, or grid cell is full
	if(!InGrid(Vec2(x,y)) || grid[x][y] != 0) {
		delete newEntity;
		return 0;
	}
	grid[x][y] = newEntity;
	entities.push_back(newEntity);
	objects.push_back(newEntity);
	return newEntity;
}

// Entity has moved positions, update the grid.
void Game::EntityMoved(Vec2 from, Vec2 to, Entity *ent) {
	grid[(int)from.x][(int)from.y] = 0;
	grid[(int)to.x][(int)to.y] = ent;
}

// Particle effect, saved for fun.
void Game::CreateExplosion(Vec2 pos) {
	for(int i = 0; i < 10; i++) {
		Particle *np = new Particle(rand() % 3 + 2, 8 + rand() % 4, Color(200 + rand() % 50, 128 + rand() % 128, 0), 3, 5, GetCenterOfCell(pos), 20);
		Game::Instance()->CreateObject(np);
	}
}

// Safe function for accessing the grid
Entity* Game::GetGridAt(Vec2 pos) {
	if(pos.x < 0 || pos.y < 0 || pos.x >= CELLS_X || pos.y >= CELLS_Y)
		return 0;
	return grid[(int)pos.x][(int)pos.y];
}

// Start a new game!
// Clears the grid and the terrain, then generates terrain (based on my arbitrary hardcoded values)
Game::Game() {
	state = GS_INGAME;
	loser = TEAM_NONE;

	// Clear grid/terrain
	gridcol = Color(50, 50, 50);
	for(int i = 0; i < CELLS_X; i++) {
		for(int j = 0; j < CELLS_Y; j++) {
			grid[i][j] = 0;
			terrain[i][j] = 0;
		}
	}

	// total number of hills (these will probably merge together)
	int numHills = 20 + rand() % 15;

	// Make hills
	for(int i = 0; i < numHills; i++) {

		// Our random hill factors
		Vec2 hillPos = Vec2(rand() % CELLS_X, rand() % CELLS_Y);	
		int hillHeight = rand() % 15 + 1 ;//rand()%4 + 2;//rand() % 6 + 2;
		int hillSize = rand() % 5 + 1;

		// The peak of the hill
		terrain[(int)hillPos.x][(int)hillPos.y] += hillHeight;
		
		// For each level of the hill, add another ring of terrain, one level lower
		// than the previous level.
		for(int j = hillHeight - 1; j > 0; j--) {
			for(int k = 0; k < hillHeight - j; k++) {
				
				// We only need to do one corner of the hill, then we can just copy and
				// invert that quadrant as needed
				Vec2 q[4];
				q[0] = Vec2(k, hillHeight - j - k);
				q[1] = Vec2(-q[0].y, q[0].x);
				q[2] = Vec2(-q[0].x, -q[0].y);
				q[3] = Vec2(-q[1].x, -q[1].y);
				
				// Actually modify the terrain
				for(int jj = 0; jj < 4; jj++) {
					Vec2 realPos = Vec2(hillPos.x + q[jj].x, hillPos.y + q[jj].y);
					if(InGrid(realPos)) {
						terrain[(int)realPos.x][(int)realPos.y] += j;
					}
				}
			}
		}

		// We're only allowing 16 depths of terrain at the moment. Let's cap stuff if it's
		// too high (or too low somehow)
		for(int i = 0; i < CELLS_X; i++) {
			for(int j = 0; j < CELLS_Y; j++) {
				if(terrain[i][j] > 15)
					terrain[i][j] = 15;
				if(terrain[i][j] < 0)
					terrain[i][j] = 0;
			}
		}
				
	}
}

// Give a global order to move to position pos to all members of the given team
void Game::OrderTeam(Team team, Vec2 pos) {
	for(Entities::iterator ei = entities.begin(); ei != entities.end(); ei++) {
		Entity *curr = (*ei);
		if(curr->team == team) {
			curr->globalOrder = pos;
		}
	}
}

// Note: This function is one of the biggest performance hogs. It'll probably be optimized
// for the next iteration. Any entities not currently fighting call this every
// frame. However, at this point things seem to still run okay, and it allows us
// to have very effective entities. 
// Iterates through the list of entities, and return the nearest ally who is
// in danger.
Vec2 Game::GetClosestFriendlyTarget(Entity *me, int maxdist) {
	float bestSqr;
	Vec2 bestTarget = Vec2(-1, -1);
	bestSqr = 99999999;
	Vec2 myPos = Vec2(me->x,me->y);
	int maxDistSqr = maxdist * maxdist;
	bool bestHasTarget = false;	// Not used atm, can be used to prioritize friendlies in combat

	for(Entities::iterator ei = entities.begin(); ei != entities.end(); ei++) {
		Entity *curr = (*ei);

		// The entity can't be the current entity, or one not on our team.
		if(curr != me && curr->team == me->team) {

			// Use sqr dist of course, since sqrt is slow (although our dist is never very high, seeing as
			// it's distance on a grid)
			float dist = myPos.distSqr(Vec2(curr->x, curr->y));

			// Base is under attack, it takes priority
			if(curr->GetType() == RT_BASE && curr->target) {
				return Vec2(curr->target->x, curr->target->y);
			}
			
			// Otherwise, let's find the nearest friendly who is currently fighting, or is investigating
			// an attack on itself.
			if((curr->target || (curr->localOrderTimeLeft > 0 && curr->lorderType == LO_ATTACKED)) 
				&& dist < bestSqr && (maxdist == 0 || dist <= maxDistSqr)){
				if(curr->target) {
					bestTarget = Vec2(curr->target->x, curr->target->y);
					bestHasTarget = true;
				}
				else {
					bestTarget = curr->localOrder;
					bestHasTarget = false;
				}
					
				bestSqr = dist;
				
			}
		}
	}
	return bestTarget;
}

// Draw *most* of our hud. (Game over message and instructions to quit)
void Game::drawHUD() {
	if(state == GS_GAMEOVER) {
	
		std::string str = "Game Over, the " + GetTeamName(loser) + " has been eliminated!"; 
		char *stringer = (char *)str.c_str();
		gfx->GFXtext_center(stringer, SC_WIDTH / 2, SC_HEIGHT / 2, Color(255,0,0), Color(0,0,0));
	}
	gfx->GFXtext_center("Push Esc To Exit", SC_WIDTH / 2, SC_HEIGHT - 2*CELL_SIZE, Color(255,255,255), Color(0,0,0));

}

// Draw the grid
void Game::drawGrid() {
	for(int i = 0; i < CELLS_X; i++) {
		gfx->GFXline(i*CELL_SIZE, 0, i*CELL_SIZE, SC_HEIGHT, gridcol);
	}
	for(int i = 0; i < CELLS_Y; i++) {
		gfx->GFXline(0, i*CELL_SIZE, SC_WIDTH, i*CELL_SIZE, gridcol);
	}
}

// Draw the terrain
void Game::drawTerrain() {
	int halfCell = (CELL_SIZE / 2);
	for(int i = 0; i < CELLS_X; i++) {
		for(int j = 0; j < CELLS_Y; j++) {
			int quality = terrain[i][j];
			if(quality != 0) {
				Vec2 center = GetCenterOfCell(Vec2(i, j));
				gfx->GFXrectfill(center.x - halfCell, center.y - halfCell, center.x + halfCell, center.y + halfCell, Color(quality*16, quality*16, quality*16)); 
			}
		}
	}
}

// Helper function that translates the enum dir into a vector.
Vec2 Game::DirToVec2(Direction dir) {

	Vec2 vTarget = Vec2(0, 0);

	switch(dir) {
	case DIR_UPLEFT:
		vTarget = Vec2(-1, -1);
		break;
	case DIR_UP:
		vTarget = Vec2(0, -1);
		break;
	case DIR_UPRIGHT:
		vTarget = Vec2(1, -1);
		break;
	case DIR_LEFT:
		vTarget = Vec2(-1, 0);
		break;
	case DIR_NONE:
		vTarget = Vec2(0, 0);
		break;
	case DIR_RIGHT:
		vTarget = Vec2(1, 0);
		break;
	case DIR_DOWNLEFT:
		vTarget = Vec2(-1, 1);
		break;
	case DIR_DOWN:
		vTarget = Vec2(0, 1);
		break;
	case DIR_DOWNRIGHT:
		vTarget = Vec2(1, 1);
		break;
	}

	return vTarget;
}

// Given a cell, returns the screen point at the center
Vec2 Game::GetCenterOfCell(Vec2 pos) {
	return Vec2(pos.x*CELL_SIZE + CELL_SIZE / 2, pos.y*CELL_SIZE + CELL_SIZE / 2);
}

// Is the given vector in the valid grid space?
bool Game::InGrid(Vec2 pos) {
	if(pos.x < 0 || pos.y < 0 || pos.x >= CELLS_X || pos.y >= CELLS_Y)
		return false;
	return true;
}

// Accessor for terrain
int Game::GetTerrainAt(Vec2 pos) {
	if(InGrid(pos))
		return terrain[(int)pos.x][(int)pos.y];
	return -1;
}

// Returns a string for the given team
// Probably could be put in a stringbank or something
std::string Game::GetTeamName(Team team) {
	std::string str = "";

	switch(team) {
	case TEAM_RED:
		str = "Red Team";
	break;
	case TEAM_BLUE:
		str = "Blue Team";
	break;
	default:
		str = "";
	break;
	}
	return str;
}

// Render everything in the game!
void Game::Render() {
	// Clear bg
	gfx->GFXclearTo(Color(0,0,0));

	// Draw the terrain
	drawTerrain();

	//if(DRAW_GRID) {
	// 	drawGrid();
	//}

	// Call render for all the renderers in the game
	for(Renderers::iterator ri = objects.begin(); ri != objects.end(); ri++) {
		(*ri)->Render();
	}
	
	// ..and finally draw our HUD.
	drawHUD();
}