/* AI WARS v0.2
	Author:	Ian Holowka
	Purpose: A program to function as a sandbox for simple swarm-type AI. The program uses the allegro engine for graphics due to it being incredibly
	simple to get working. However, the engine has been designed to support multiple different graphics implementations (using the GFX class).
	The "game" involves a number of AI-controlled geometrical shapes fighting to destroy each other's bases.
	There are (at this point) three main units:
		1) Soldier (triangle), a basic infantry unit with short range
		2) Tank (square), a powerful but slow unit with medium range capable of doing splash damage
		3) Sniper (triangle with plus), similar to soldier, but VERY slow, with very long range and poor health
	The units follow a very simple logic of:
		1) Fighting any units within range
		2) Assisting any friends who are in danger, or investigating shots fired at them
		3) Moving towards their global goal (enemy base).
	The logic was designed to be simple to allow for many units to be in the game at any time.
	There is also terrain. The brighter the terrain, the better the cover provided (damage redux). Units will attempt to move to
	better cover wherever possible.
	The main class sets up the game and global orders. It also, in this case, implements a reinforcement system where
	each team are given half of their forces back at the end of a given time period.

	COMMAND LINE ARGUMENTS: 
		Please examine the included batch files for information on how to modify the amount of troops on each team.

	Note: Comments are mainly in the .cpp files.
*/

#include <stdlib.h>
#include <sstream>
#include "Game.h"
#include "Entities.h"

#define DRAW_GRID 1
#define REINFORCEMENTS_TIME 3000

bool gameOver = false;

volatile int ticks = 0;
Game *game;
GFX *gfx;
int blue_soldiers;
int blue_tanks;
int blue_snipers;

int red_soldiers;
int red_tanks;
int red_snipers;

int	screenh;
int screenv;
GFXmode windowed;

int reinforcements;

// Timing callback
void ticker()
{
        ticks++;
}
#ifdef BUILD_ALLEGRO
END_OF_FUNCTION(ticker)
#endif 


void init_game() {

	// Random seed!
	srand(time(0));
	
	// Init our singletons
	gfx = GFX::Instance();
	game = Game::Instance();
	
	// Setup our timing callback
	LOCK_VARIABLE(ticks);
    LOCK_FUNCTION(ticker);
    install_int_ex(ticker, BPS_TO_TIMER(FPS));

	reinforcements = REINFORCEMENTS_TIME;

	// Create the bases
	game->AddEntity(new Base(CELLS_X / 2, 3, 1000, TEAM_RED));
	game->AddEntity(new Base(CELLS_X / 2, CELLS_Y - 3, 1000, TEAM_BLUE));

	int b1;
	int b2;
	int b3;
	int c1;
	int highest;

	// Find out which unit number is the highest.
	if(red_soldiers > blue_soldiers)	b1 = red_soldiers;
	else								b1 = blue_soldiers;
	if(red_tanks > blue_tanks)			b2 = red_soldiers;
	else								b2 = blue_tanks;
	if(red_snipers > blue_snipers)		b3 = red_snipers;
	else								b3 = blue_snipers;
	if(b1 > b2)							c1 = b1;
	else								c1 = b2;
	if(b3 > c1)							highest = b3;
	else								highest = c1;

	int timeout;

	// Create our units, mixed together.
	for(int i = 0; i < highest; i++) {
		if(i < red_soldiers) {
			timeout = 10;
			Entity *curr = 0;
			while(!curr && timeout-- > 0)
				curr = game->AddEntity(new Soldier(rand() % CELLS_X + 1, rand() % (int)(CELLS_Y/3) + 1, SOLDIER_GRUNT, TEAM_RED));
		}
		if(i < blue_soldiers) {
			timeout = 10;
			Entity *curr = 0;
			while(!curr && timeout-- > 0)
				curr = game->AddEntity(new Soldier(rand() % CELLS_X + 1, CELLS_Y - (rand() % (int)(CELLS_Y/3) + 1), SOLDIER_GRUNT, TEAM_BLUE));
		}
		if(i < red_tanks) {
			timeout = 10;
			Entity *curr = 0;
			while(!curr && timeout-- > 0)
				curr = game->AddEntity(new Tank(rand() % CELLS_X + 1, rand() % (int)(CELLS_Y/3) + 1,  250, 10 + rand() % 10 , rand() % 3 + 3, Color(128 + rand() % 128, 0, 0), TEAM_RED));
		}
		if(i < blue_tanks) {
			timeout = 10;
			Entity *curr = 0;
			while(!curr && timeout-- > 0)
				curr = game->AddEntity(new Tank(rand() % CELLS_X + 1, CELLS_Y - (rand() % (int)(CELLS_Y/3) + 1),  250, 10 + rand() % 10, rand() % 3 + 3, Color(0, 0, 128 + rand() % 128), TEAM_BLUE));
		}
		if(i < red_snipers) {
			timeout = 10;
			Entity *curr = 0;
			while(!curr && timeout-- > 0)
				curr = game->AddEntity(new Soldier(rand() % CELLS_X + 1, rand() % (int)(CELLS_Y/3) + 1, SOLDIER_SNIPER, TEAM_RED));
		}
		if(i < blue_snipers) {
			timeout = 10;
			Entity *curr = 0;
			while(!curr && timeout-- > 0)
				curr = game->AddEntity(new Soldier(rand() % CELLS_X + 1, CELLS_Y - (rand() % (int)(CELLS_Y/3) + 1), SOLDIER_SNIPER, TEAM_BLUE));
		}
	}

	// Give out our team orders.
	game->OrderTeam(TEAM_RED, Vec2(CELLS_X / 2, CELLS_Y - 3));
	game->OrderTeam(TEAM_BLUE, Vec2(CELLS_X / 2, 3));
}

// Process command line input
void process_args(int argc, char *argv[]) {

	// Commented stuff is for resolution changing via command line (not yet impl)
/*	screenh = 800;
	screenv = 600;
	windowed = GFXMODE_FULLSCREEN;

	if(argc >= 3) {
		int twindowed = atoi(argv[1]);		// 0 for fullscreen, 1 for windowed
		int tscreenh = atoi(argv[2]);
		int tscreenv = atoi(argv[3]);
		if(twindowed < 2) {
			if(	(tscreenh == 800 && tscreenv == 600) ||
				(tscreenh == 1024 && tscreenv == 768) ||
				(tscreenh == 1280 && tscreenv == 1024) ||
				(tscreenh == 1600 && tscreenv == 1200) ||
				(tscreenh == 1650 && tscreenv == 1080) ) {
					screenh = tscreenh;
					screenv = tscreenv;
					if(twindowed == 1)	windowed = GFXMODE_WINDOW;
					else				windowed = GFXMODE_FULLSCREEN;
			}
		}
	}*/
	
	// Default
	if(argc < 6) {
		blue_soldiers = 200;
		blue_tanks = 50;
		blue_snipers = 50;
		red_soldiers = 200;
		red_tanks = 50;
		red_snipers = 50;
	}
	else {
		blue_soldiers = atoi(argv[1]);
		blue_tanks = atoi(argv[2]);
		blue_snipers = atoi(argv[3]);
		red_soldiers = atoi(argv[4]);
		red_tanks = atoi(argv[5]);
		red_snipers = atoi(argv[6]);
	}
}

// Call in reinforcements
void DoReinforcements() {
	reinforcements = REINFORCEMENTS_TIME;

	for(int i = 0; i < red_soldiers / 2; i++){
		game->AddEntity(new Soldier(rand() % CELLS_X + 1, rand() % (int)(CELLS_Y/3) + 1, SOLDIER_GRUNT, TEAM_RED));
	} 

	for(int i = 0; i < blue_soldiers / 2; i++) {
		game->AddEntity(new Soldier(rand() % CELLS_X + 1, CELLS_Y - (rand() % (int)(CELLS_Y/3) + 1), SOLDIER_GRUNT, TEAM_BLUE));
	}

	for(int i = 0; i < blue_snipers / 2; i++) {
		game->AddEntity(new Soldier(rand() % CELLS_X + 1, CELLS_Y - (rand() % (int)(CELLS_Y/3) + 1), SOLDIER_SNIPER, TEAM_BLUE));
	}

	for(int i = 0; i < red_snipers / 2; i++){
		game->AddEntity(new Soldier(rand() % CELLS_X + 1, rand() % (int)(CELLS_Y/3) + 1, SOLDIER_SNIPER, TEAM_RED));
	}

	for(int i = 0; i < red_tanks / 2; i++){
		game->AddEntity(new Tank(rand() % CELLS_X + 1, rand() % (int)(CELLS_Y/3) + 1,  250, 10 + rand() % 10 , rand() % 3 + 3, Color(128 + rand() % 128, 0, 0), TEAM_RED));
	} 

	for(int i = 0; i < blue_tanks / 2; i++) {
		game->AddEntity(new Tank(rand() % CELLS_X + 1, CELLS_Y - (rand() % (int)(CELLS_Y/3) + 1),  250, 10 + rand() % 10, rand() % 3 + 3, Color(0, 0, 128 + rand() % 128), TEAM_BLUE));
	}

	game->OrderTeam(TEAM_RED, Vec2(CELLS_X / 2, CELLS_Y - 3));
	game->OrderTeam(TEAM_BLUE, Vec2(CELLS_X / 2, 3));
}

// Program entry point
int main(int argc, char *argv[]) 
{
	process_args(argc, argv);

#ifdef BUILD_ALLEGRO
	allegro_init();
	install_keyboard();
#endif

	init_game();
	
	// Our main game loop
	while(!key[KEY_ESC]) {
		
		// Wait for update
		while(ticks == 0)
		{
			rest(100 / FPS);
		}

		// Update for the ticks that have gone by
		while(ticks > 0)
		{
			int old_ticks = ticks;

			game->Update();

			ticks--;
			if(old_ticks <= ticks)
				break;
		}

		// Reinforce
		if(!game->state == GS_GAMEOVER) {
			reinforcements--;
			if(reinforcements <= 0) {
				DoReinforcements();
			}
		}

		// Our main render call
		game->Render();

		// Typically this would be in game, however, the "reinforcements" is managed entirely by
		// the main program.
		std::stringstream reinf;
		reinf << reinforcements;
		std::string str = "Reinforcements in " + reinf.str();
		char *stringer = (char *)str.c_str();
		gfx->GFXtext(stringer, SC_WIDTH / 2, 2*CELL_SIZE, Color(255, 255, 255), Color(0,0,0));

		// Swap buffers
		gfx->GFXupdate();


	}

	// Clean up gfx
	delete(gfx);
	gfx = 0;

	return 0;
}
#ifdef BUILD_ALLEGRO
END_OF_MAIN()
#endif
